# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import logging
import threading
import uuid
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    get_type_hints,
)

from pydantic import BaseModel, Field, PrivateAttr

if TYPE_CHECKING:
    from .dispatcher import ProvenanceChain
    from .graph import Graph


class Task(BaseModel):
    _provenance: List[Tuple[str, int]] = PrivateAttr(default_factory=list)
    _input_provenance: List["Task"] = PrivateAttr(default_factory=list)
    _retry_count: int = PrivateAttr(default=0)

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def retry_count(self) -> int:
        """
        Read-only property to access the current retry count.
        """
        return self._retry_count

    def increment_retry_count(self) -> None:
        """
        Increments the retry count by 1.
        """
        self._retry_count += 1
        return True

    def copy_provenance(self) -> List[Tuple[str, int]]:
        return self._provenance.copy()

    def copy_input_provenance(self) -> List["Task"]:
        return [input.model_copy() for input in self._input_provenance]

    def find_input_task(self, task_class: Type["Task"]) -> Optional["Task"]:
        """
        Find the most recent input task of the specified class in the input provenance.

        Args:
            task_class (Type[Task]): The class of the task to find.

        Returns:
            Optional[Task]: The most recent task of the specified class,
                                    or None if no such task is found.
        """
        for task in reversed(self._input_provenance):
            if task.__class__ is task_class:
                return task
        return None

    def previous_input_task(self):
        return self._input_provenance[-1] if self._input_provenance else None

    def prefix_for_input_task(
        self, task_class: Type["Task"]
    ) -> Optional["ProvenanceChain"]:
        """
        Finds the provenance chain for the most recent input task of the specified class.

        Args:
            task_class (Type[Task]): The class of the task to find.
        Returns:
            ProvenanceChain: The provenance chain for the most recent input task of the specified class.
        """
        for i in range(len(self._provenance) - 1, -1, -1):
            if self._provenance[i][0] == task_class.__name__:
                return tuple(self._provenance[: i + 1])
        return None


class WorkBufferContext:
    def __init__(self, worker, input_task=None):
        self.worker: "TaskWorker" = worker
        self.input_task: "Task" = input_task
        self.work_buffer: List[Tuple["TaskWorker", "Task"]] = []

    def __enter__(self):
        self.worker._local.ctx = self
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._flush_work_buffer()
        self.worker._local.ctx = None

    def get_input_and_outputs(self):
        return self.input_task, [entry[1] for entry in self.work_buffer]

    def _flush_work_buffer(self):
        if self.worker._graph and self.worker._graph._dispatcher:
            logging.info(
                "Worker %s flushing work buffer with %d items",
                self.worker.name,
                len(self.work_buffer),
            )
            self.worker._graph._dispatcher.add_multiple_work(self.work_buffer)
        else:
            for consumer, task in self.work_buffer:
                self.worker._dispatch_work(task)
        self.work_buffer.clear()

    def add_to_buffer(self, consumer: "TaskWorker", task: "Task"):
        self.work_buffer.append((consumer, task))


class TaskWorker(BaseModel, ABC):
    """
    Base class for all task workers.

    This class is strongly typed for both input and output types. The type checking
    is performed during the registration of consumers.

    Attributes:
        output_types (Set[Type[Task]]): The types of work this task can output.
        num_retries (int): The number of retries allowed for this task. Defaults to 0.
        _id (int): A private attribute to track the task's ID.
        _consumers (Dict[Type["TaskWorker"], "TaskWorker"]): A private attribute to store registered consumers.

    Note:
        Any subclass of TaskWorker must implement consume_work.
    """

    output_types: Set[Type[Task]] = Field(
        default_factory=set,
        description="The types of work this task can output",
    )
    num_retries: int = Field(
        0, description="The number of retries allowed for this task"
    )

    _state_lock: threading.Lock = PrivateAttr(default_factory=threading.Lock)
    _id: int = PrivateAttr(default=0)
    _consumers: Dict[Type["TaskWorker"], "TaskWorker"] = PrivateAttr(
        default_factory=dict
    )
    _graph: Optional["Graph"] = PrivateAttr(default=None)
    _last_input_task: Optional[Task] = PrivateAttr(default=None)
    _instance_id: uuid.UUID = PrivateAttr(default_factory=uuid.uuid4)
    _local: threading.local = PrivateAttr(default_factory=threading.local)

    def __init__(self, **data):
        super().__init__(**data)

    def __hash__(self):
        return hash(self._instance_id)

    def __eq__(self, other):
        if isinstance(other, TaskWorker):
            return self._instance_id == other._instance_id
        return False

    def work_buffer_context(self, input_task):
        if input_task is None:
            raise ValueError("Input task cannot be None")
        return WorkBufferContext(self, input_task)

    @property
    def name(self) -> str:
        """
        Returns the name of this worker class.

        :return: The name of the class.
        :rtype: str
        """
        return self.__class__.__name__

    @property
    def last_input_task(self) -> Optional[Task]:
        """
        Returns the last input task consumed by this worker.

        :return: The last input task as a Task object, or None if there is no last input task.
        :rtype: Optional[Task]
        """
        with self._state_lock:
            return self._last_input_task

    def set_graph(self, graph: "Graph"):
        self._graph = graph

    def next(self, downstream: "TaskWorker"):
        """
        Sets the dependency between the current task and the downstream task.

        Parameters:
            downstream (TaskWorker): The downstream task to set as a dependency.

        Returns:
            TaskWorker: The downstream task.

        Raises:
            ValueError: If the task has not been added to a Graph before setting dependencies.
        """
        if self._graph is None:
            raise ValueError(
                "Task must be added to a Graph before setting dependencies"
            )
        self._graph.set_dependency(self, downstream)
        return downstream

    def trace(self, prefix: "ProvenanceChain"):
        """
        Traces the provenance chain for a given prefix in the graph.

        This method sets up a trace on a given prefix in the provenance chain. It will be visible
        in the dispatcher dashboard.

        Parameters:
        -----------
        prefix : ProvenanceChain
            The prefix to trace. Must be a tuple representing a part of a task's provenance chain.
            This is the sequence of task identifiers leading up to (but not including) the current task.
        """
        self._graph._dispatcher.trace(prefix)

    def watch(self, prefix: "ProvenanceChain", task: Optional[Task] = None) -> bool:
        """
        Watches for the completion of a specific provenance chain prefix in the task graph.

        This method sets up a watch on a given prefix in the provenance chain. It will be notified
        in its notify method when this prefix is no longer part of any active task's provenance, indicating
        that all tasks with this prefix have been completed.

        Parameters:
        -----------
        prefix : ProvenanceChain
            The prefix to watch. Must be a tuple representing a part of a task's provenance chain.
            This is the sequence of task identifiers leading up to (but not including) the current task.

        task : Optional[Task], default=None
            The task associated with this watch operation. This parameter is optional and may be
            used for additional context or functionality in the underlying implementation.

        Returns:
        --------
        bool
            True if the watch was successfully added for the given prefix.
            False if a watch for this prefix was already present.

        Raises:
        -------
        ValueError
            If the provided prefix is not a tuple.
        """
        if not isinstance(prefix, tuple):
            raise ValueError("Prefix must be a tuple")
        return self._graph._dispatcher.watch(prefix, self, task)

    def unwatch(self, prefix: "ProvenanceChain") -> bool:
        """
        Removes the watch for this task provenance to be completed in the graph.

        Parameters:
            worker (Type["Task"]): The worker to unwatch.

        Returns:
            True if the watch was removed, False if the watch was not present.
        """
        if not isinstance(prefix, tuple):
            raise ValueError("Prefix must be a tuple")
        return self._graph._dispatcher.unwatch(prefix, self)

    def _pre_consume_work(self, task: Task):
        with self._state_lock:
            self._last_input_task = task
        with self.work_buffer_context(task):
            self.consume_work(task)

    def init(self):
        """
        Called when the graph is fully constructed and starts work.
        """
        pass

    @abstractmethod
    def consume_work(self, task: Task):
        """
        Abstract method to consume a work item.

        This method must be implemented by subclasses to define specific work consumption logic. It needs to be thread-safe
        as it may be called concurrently by multiple threads.

        Args:
            task (Task): The work item to be consumed.
        """
        pass

    def publish_work(self, task: Task, input_task: Optional[Task]):
        """
        Publish a work item.

        This method handles the publishing of work items, including provenance tracking and consumer routing.

        Args:
            task (Task): The work item to be published.
            input_task (Task): The input task that led to this work item.

        Raises:
            ValueError: If the task type is not in the output_types or if no consumer is registered for the task type.
        """
        if type(task) not in self.output_types:
            raise ValueError(
                f"Task {self.name} cannot publish work of type {type(task).__name__}"
            )

        # Copy provenance from input task if provided
        if input_task is not None:
            task._provenance = input_task.copy_provenance()
            task._input_provenance = input_task.copy_input_provenance() + [
                input_task.model_copy()
            ]
        else:
            task._provenance = []
            task._input_provenance = []

        with self._state_lock:
            self._id += 1

        task._provenance.append((self.name, self._id))
        logging.info(
            "Task %s published work with provenance %s", self.name, task._provenance
        )

        # Verify if there is a consumer for the given task class
        consumer = self._consumers.get(task.__class__)
        if consumer is None:
            raise ValueError(f"No consumer registered for {task.__class__.__name__}")

        logging.info(
            "Worker %s publishing work to consumer %s with task type %s",
            self.name,
            consumer.name,
            task.__class__.__name__,
        )

        # this requires that anything that might call publish_work is wrapped in a work_buffer_context
        self._local.ctx.add_to_buffer(consumer, task)

    def completed(self):
        """Called to let the worker know that it has finished processing all work."""
        pass

    def notify(self, task_name: str):
        """Called to notify the worker that no tasks with provenance of task_name are remaining."""
        pass

    def _dispatch_work(self, task: Task):
        consumer = self._consumers.get(task.__class__)
        consumer.consume_work(task)

    def validate_task(
        self, task_cls: Type[Task], consumer: "TaskWorker"
    ) -> Tuple[bool, Exception]:
        """
        Validate that a consumer can handle a specific Task type.

        This method checks if the consumer has a properly typed consume_work method for the given task class.

        Args:
            task_cls (Type[Task]): The Task subclass to validate.
            consumer (TaskWorker): The consumer to validate against.

        Returns:
            Tuple[bool, Exception]: A tuple containing a boolean indicating success and an exception if validation failed.
        """
        # Ensure consumer has a consume_work method taking task_cls as parameter
        consume_method = getattr(consumer, "consume_work", None)
        if not consume_method:
            return False, AttributeError(
                f"{consumer.__class__.__name__} has no method consume_work"
            )

        # Check the type of the first parameter of consume_work method
        signature = inspect.signature(consume_method)
        parameters = signature.parameters
        if len(parameters) != 1:
            return False, TypeError(
                f"Method consume_work in {consumer.__class__.__name__} must accept one parameter"
            )

        # Get the type hints for consume_work
        type_hints = get_type_hints(consume_method)
        first_param_type = type_hints.get("task", None)

        if first_param_type is not task_cls:
            return False, TypeError(
                f"TaskWorker {consumer.__class__.__name__} cannot consume tasks of type {task_cls.__name__}. It can only consume tasks of type {first_param_type.__name__}"
            )

        return True, None

    def get_task_class(self) -> Type[Task]:
        """
        Get the Task subclass that this worker can consume.

        This method inspects the consume_work method to determine the type of Task it can handle.

        Returns:
            Type[Task]: The Task subclass this worker can consume.

        Raises:
            AttributeError: If the consume_work method is not defined.
            TypeError: If the consume_work method is not properly typed.
        """
        consume_method = getattr(self, "consume_work", None)
        if not consume_method:
            raise AttributeError(
                f"{self.__class__.__name__} has no method consume_work"
            )
        signature = inspect.signature(consume_method)
        parameters = signature.parameters
        if len(parameters) != 1:
            raise TypeError(
                f"Method consume_work in {self.__class__.__name__} must accept one parameter"
            )
        type_hints = get_type_hints(consume_method)
        first_param_type = type_hints.get("task", None)
        if not first_param_type:
            raise TypeError(
                f"consume_work method in {self.__class__.__name__} must have type hints"
            )
        return first_param_type

    def register_consumer(self, task_cls: Type[Task], consumer: "TaskWorker"):
        """
        Register a consumer for a specific Task type.

        This method performs type checking to ensure that the consumer can handle the specified Task type.

        Args:
            task_cls (Type[Task]): The Task subclass to register a consumer for.
            consumer (TaskWorker): The consumer to register.

        Raises:
            TypeError: If task_cls is not a subclass of Task or if the consumer cannot handle the task type.
            ValueError: If the task type is not in the output_types or if a consumer is already registered for the task type.
        """
        # Ensure task_cls is a subclass of Task
        if not issubclass(task_cls, Task):
            raise TypeError(f"{task_cls.__name__} is not a subclass of Task")

        success, error = self.validate_task(task_cls, consumer)
        if not success:
            raise error

        if task_cls not in self.output_types:
            raise ValueError(
                f"Downstream consumer {consumer.name} only accepts work of type {task_cls.__name__} but Worker {self.name} does not produce it"
            )

        if task_cls in self._consumers:
            raise ValueError(f"Consumer for {task_cls.__name__} already registered")
        self._consumers[task_cls] = consumer

        # special cases like JoinedTask need to validate that their join_type is upstream
        consumer._validate_connection()

    def _validate_connection(self):
        pass


def main():
    class MagicTaskWork(Task):
        magic: Any = Field(..., title="Magic", description="Magic value")

    class SpecificTask(TaskWorker):
        output_types: Set[Type[Task]] = {MagicTaskWork}

        def consume_work(self, task: MagicTaskWork):
            print(
                f"SpecificTask {self.name} consuming work item with provenance: {task._provenance}: {task.magic}"
            )

    # Example usage
    task_a = SpecificTask(name="TaskA")
    work_item = MagicTaskWork(magic="something")

    # Register a consumer (for demo purposes, registering itself)
    task_a.register_consumer(MagicTaskWork, task_a)

    # Publish work
    task_a.publish_work(work_item, input_task=None)


if __name__ == "__main__":
    main()



# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import logging
import os
import re
from typing import Any, Dict, Literal, Optional, Tuple

import diskcache
from dotenv import load_dotenv
from langchain_core.output_parsers import PydanticOutputParser
from langchain_core.prompts import PromptTemplate
from langchain_core.pydantic_v1 import BaseModel
from ollama import Client

from .openai import OpenAIWrapper
from .remote_ollama import RemoteOllama
from .ssh import SSHConnection
from .utils import setup_logging

# Load environment variables from .env.local file
load_dotenv(".env.local")


class LLMInterface:
    def __init__(
        self,
        model_name: str = "llama2",
        log_dir: str = "logs",
        client: Optional[Any] = None,
    ):
        self.model_name = model_name
        self.client = client if client else Client()

        self.logger = setup_logging(
            logs_dir=log_dir, logs_prefix="llm_interface", logger_name=__name__
        )

        # Initialize disk cache for caching responses
        self.disk_cache = diskcache.Cache(
            directory=".response_cache", eviction_policy="least-recently-used"
        )

    def _cached_generate(self, prompt: str, system: str = "", format: str = "") -> str:
        # Hash the prompt to use as the cache key
        prompt_hash = self._generate_hash(
            self.model_name + "\n" + system + "\n" + prompt
        )

        # Check if prompt response is in cache
        response = self.disk_cache.get(prompt_hash)

        if response is None:
            # If not in cache, make request to client
            response = self.client.generate(
                model=self.model_name, prompt=prompt, system=system, format=format
            )

            # Cache the response with hashed prompt as key
            self.disk_cache.set(prompt_hash, response)

        return response

    def generate(self, prompt: str, system: str = "") -> str:
        self.logger.info("Generating text with prompt: %s...", prompt[:850])
        response = self._cached_generate(prompt=prompt, system=system)
        self.logger.info("Generated text: %s...", response["response"][:850])
        return response["response"].strip()

    def _strip_text_from_json_response(self, response: str) -> str:
        pattern = r"^[^{\[]*([{\[].*[}\]])[^}\]]*$"
        match = re.search(pattern, response, re.DOTALL)

        if match:
            return match.group(1)
        else:
            return response  # Return original response if no JSON block is found

    def generate_pydantic(
        self,
        prompt_template: str,
        output_schema: BaseModel,
        system: str = "",
        logger: Optional[logging.Logger] = None,
        **kwargs,
    ) -> Optional[BaseModel]:
        """
        Generate a python object based on the given prompt template and output schema.

        Args:
            prompt_template (str): The prompt template with placeholders for variables.
            output_schema (BaseModel): A Pydantic model defining the expected output structure.
            system (str): The system prompt to use for generation.
            **kwargs: Keyword arguments to fill in the prompt template.

        Returns:
            Dict[str, Any]: The formatted prompt and generated output in JSON format.
        """
        parser = PydanticOutputParser(pydantic_object=output_schema)

        prompt = PromptTemplate(
            template=prompt_template,
            input_variables=list(kwargs.keys()),
        )

        formatted_prompt = prompt.format(**kwargs)

        self.logger.info("Generated prompt: %s", formatted_prompt)
        if logger:
            logger.info("Generated prompt: %s", formatted_prompt)

        iteration = 0
        while iteration < 3:
            iteration += 1

            raw_response = self._cached_generate(
                prompt=formatted_prompt, system=system, format="json"
            )["response"].strip()

            error_message, response = self._parse_response(raw_response, parser)

            if response is not None:
                break

            formatted_prompt = (
                f"Your previous response did not follow the JSON format instructions. Here is the error message {error_message}\n\n"
                f"Try again and closly follow these instructions:\n{formatted_prompt}"
            )

        return response

    def _generate_hash(self, prompt: str) -> str:
        hash_object = hashlib.sha256(prompt.encode())
        return hash_object.hexdigest()

    def _parse_response(
        self, response: str, parser: PydanticOutputParser
    ) -> Tuple[str, Dict[str, Any]]:
        self.logger.info("Parsing JSON response: %s", response)
        error_message = None
        try:
            response = parser.parse(response)
        except Exception as e:
            self.logger.error("Error parsing response: %s", e)
            error_message = str(e)
            response = None
        return error_message, response


def llm_from_config(
    provider: Literal["ollama", "remote_ollama", "openai"] = "ollama",
    model_name: str = "llama3",
    hostname: Optional[str] = None,
    username: Optional[str] = None,
    log_dir: str = "logs",
) -> LLMInterface:
    match provider:
        case "openai":
            api_key = os.getenv("OPENAI_API_KEY")
            if api_key is None:
                raise ValueError("OPENAI_API_KEY not found in environment variables")
            wrapper = OpenAIWrapper(api_key=api_key)
            return LLMInterface(
                model_name=model_name,
                log_dir=log_dir,
                client=wrapper,
            )
        case "remote_ollama":
            ssh = SSHConnection(
                hostname=hostname,
                username=username,
            )
            client = RemoteOllama(ssh_connection=ssh, model_name=model_name)
            return LLMInterface(
                model_name=model_name,
                log_dir=log_dir,
                client=client,
            )
        case "ollama":
            return LLMInterface(
                model_name=model_name,
                log_dir=log_dir,
            )

    raise ValueError(f"Invalid LLM provider in config: {provider}")



# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# src/planai/__init__.py

from .cached_task import CachedTaskWorker
from .graph import Graph
from .joined_task import JoinedTaskWorker
from .llm_interface import llm_from_config
from .llm_task import CachedLLMTaskWorker, LLMTaskWorker
from .task import Task, TaskWorker
from .utils import PydanticDictWrapper

# If you want to control what gets imported with "from planai import *"
__all__ = [
    "Graph",
    "llm_from_config",
    "LLMTaskWorker",
    "TaskWorker",
    "Task",
    "CachedTaskWorker",
    "CachedLLMTaskWorker",
    "JoinedTaskWorker",
    "PydanticDictWrapper",
]



# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import select
import threading

import paramiko

from .utils import setup_logging

try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer

logger = setup_logging(logs_prefix="ssh_connection", logger_name=__name__)


class SSHConnection:
    def __init__(self, hostname, username, port=22):
        self.hostname = hostname
        self.username = username
        self.port = port
        self.client = None
        self.forward_thread = None

    def is_active(self):
        return (
            self.client is not None
            and self.client.get_transport() is not None
            and self.client.get_transport().is_active()
        )

    def connect(self):
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        # Get keys from the SSH agent
        agent = paramiko.Agent()
        keys = agent.get_keys()
        if not keys:
            logger.error("No keys loaded into SSH agent")
            raise Exception("No keys loaded into SSH agent")

        # Try connecting with each key
        for key in keys:
            try:
                self.client.connect(
                    self.hostname, port=self.port, username=self.username, pkey=key
                )
                logger.info(
                    "Successfully connected to {} as {}".format(
                        self.hostname, self.username
                    )
                )
                return  # Exit the method if connection is successful
            except paramiko.AuthenticationException as auth_error:
                logger.warning(
                    "Authentication with key {} failed: {}".format(
                        key.get_base64(), auth_error
                    )
                )

        logger.error("All agent keys failed to authenticate")
        raise Exception("All agent keys failed to authenticate")

    def start_port_forward(self, local_port, remote_host, remote_port):
        class ForwardServer(SocketServer.ThreadingTCPServer):
            daemon_threads = True
            allow_reuse_address = True

        class Handler(SocketServer.BaseRequestHandler):
            def handle(self):
                try:
                    chan = self.ssh_transport.open_channel(
                        "direct-tcpip",
                        (self.chain_host, self.chain_port),
                        self.request.getpeername(),
                    )
                except Exception as e:
                    logger.error(
                        "Incoming request to {}:{} failed: {}".format(
                            self.chain_host, self.chain_port, repr(e)
                        )
                    )
                    return
                if chan is None:
                    logger.error(
                        "Incoming request to {}:{} was rejected by the SSH server.".format(
                            self.chain_host, self.chain_port
                        )
                    )
                    return

                logger.info(
                    "Connected! Tunnel open {} -> {} -> {}".format(
                        self.request.getpeername(),
                        chan.getpeername(),
                        (self.chain_host, self.chain_port),
                    )
                )
                while True:
                    r, w, x = select.select([self.request, chan], [], [])
                    if self.request in r:
                        data = self.request.recv(1024)
                        if len(data) == 0:
                            break
                        chan.send(data)
                    if chan in r:
                        data = chan.recv(1024)
                        if len(data) == 0:
                            break
                        self.request.send(data)

                peername = self.request.getpeername()
                chan.close()
                self.request.close()
                logger.info("Tunnel closed from {}".format(peername))

        def forward_tunnel():
            ForwardServer(("", local_port), Handler).serve_forever()

        Handler.chain_host = remote_host
        Handler.chain_port = remote_port
        Handler.ssh_transport = self.client.get_transport()

        self.forward_thread = threading.Thread(target=forward_tunnel)
        self.forward_thread.daemon = True
        self.forward_thread.start()

    def execute_command(self, command):
        if not self.client:
            logger.error("Not connected. Call connect() first.")
            raise Exception("Not connected. Call connect() first.")
        stdin, stdout, stderr = self.client.exec_command(command)
        exit_status = stdout.channel.recv_exit_status()
        return {
            "exit_status": exit_status,
            "stdout": stdout.read().decode("utf-8"),
            "stderr": stderr.read().decode("utf-8"),
        }

    def stop_port_forward(self):
        if self.forward_thread:
            transport = self.client.get_transport()
            if transport:
                transport.cancel_port_forward("", 0)
            self.forward_thread.join()
            self.forward_thread = None

    def close(self):
        if self.client:
            self.client.close()
            self.client = None



# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import threading
import time
from typing import TYPE_CHECKING, Optional

from flask import Flask, Response, jsonify, render_template, send_from_directory

if TYPE_CHECKING:
    from .dispatcher import Dispatcher

app = Flask(__name__, template_folder="templates", static_folder="static")

dispatcher: Optional["Dispatcher"] = None
quit_event = threading.Event()


@app.route("/")
def index():
    return render_template("index.html")


@app.route("/static/<path:path>")
def send_static(path):
    return send_from_directory("static", path)


@app.route("/stream")
def stream():
    def event_stream():
        last_data = None
        while True:
            current_data = get_current_data()
            if current_data != last_data:
                yield f"data: {json.dumps(current_data)}\n\n"
            last_data = current_data
            time.sleep(0.2)

    def get_current_data():
        queued_tasks = dispatcher.get_queued_tasks()
        active_tasks = dispatcher.get_active_tasks()
        completed_tasks = dispatcher.get_completed_tasks()
        failed_tasks = dispatcher.get_failed_tasks()

        data = {
            "queued": queued_tasks,
            "active": active_tasks,
            "completed": completed_tasks,
            "failed": failed_tasks,
        }

        return data

    return Response(event_stream(), mimetype="text/event-stream")


@app.route("/trace_stream")
def trace_stream():
    def event_stream():
        last_trace = None
        while True:
            current_trace = dispatcher.get_traces()
            if current_trace != last_trace:
                # Convert tuple keys to strings
                serializable_trace = {
                    "_".join(map(str, k)): v for k, v in current_trace.items()
                }
                yield f"data: {json.dumps(serializable_trace)}\n\n"
            last_trace = current_trace
            time.sleep(0.2)

    return Response(event_stream(), mimetype="text/event-stream")


@app.route("/quit", methods=["POST"])
def quit():
    global quit_event
    quit_event.set()
    return jsonify({"status": "ok"})


def run_web_interface(disp: "Dispatcher", port=5000):
    global dispatcher
    dispatcher = disp
    app.run(debug=False, port=port)


def is_quit_requested():
    global quit_event
    return quit_event.is_set()



# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Literal, Mapping

from openai import OpenAI


class OpenAIWrapper:
    def __init__(self, api_key: str):
        self.client = OpenAI(api_key=api_key)

    def generate(
        self,
        prompt: str,
        system: str = "",
        format: Literal["", "json"] = "",
        model: str = "gpt-3.5-turbo",
    ) -> Mapping[str, Any]:
        """
        Create a response using the requested OpenAI model.

        Args:
            prompt (str): The main input prompt for the model.
            system (str, optional): The system message to set the behavior of the assistant. Defaults to ''.
            format (Literal['', 'json'], optional): If set to 'json', the response will be in JSON format. Defaults to ''.
            model (str, optional): The OpenAI model to use. Defaults to 'gpt-3.5-turbo'.

        Raises:
            Exception: For any API errors.

        Returns:
            Mapping[str, Any]: A dictionary containing the response and completion status.
        """
        messages = []

        if system:
            messages.append({"role": "system", "content": system})
        elif format == "json":
            messages.append(
                {
                    "role": "system",
                    "content": "You are a helpful assistant that responds in JSON format.",
                }
            )

        messages.append({"role": "user", "content": prompt})

        api_params = {
            "model": model,
            "messages": messages,
        }

        if format == "json":
            api_params["response_format"] = {"type": "json_object"}

        try:
            response = self.client.chat.completions.create(**api_params)
            content = response.choices[0].message.content

            return {"response": content, "done": True}

        except Exception as e:
            raise e

    def chat(self, messages: list[Mapping[str, str]], **kwargs):
        """
        Conduct a chat conversation using the OpenAI API.

        Args:
            messages (list[Mapping[str, str]]): A list of message dictionaries, each containing 'role' and 'content'.
            **kwargs: Additional arguments to pass to the generate function.

        Returns:
            The response from the generate function.
        """
        system_message = next(
            (m["content"] for m in messages if m["role"] == "system"), ""
        )
        user_messages = [m["content"] for m in messages if m["role"] == "user"]

        prompt = "\n".join(user_messages)

        return self.generate(prompt=prompt, system=system_message, **kwargs)



# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import os
import time
from contextlib import contextmanager
from datetime import datetime
from typing import Any, Dict, Optional

from pydantic import BaseModel


def setup_logging(
    logs_dir: str = "logs",
    logs_prefix="general_log",
    logger_name: Optional[str] = None,
    level: Optional[int] = logging.INFO,
) -> logging.Logger:
    if not os.path.exists(logs_dir):
        os.makedirs(logs_dir)

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    log_file = os.path.join(logs_dir, f"{logs_prefix}_{timestamp}.log")

    logger = logging.getLogger(name=logger_name)
    logger.handlers = []
    logger.setLevel(level)

    # Create file handler but don't create the file until the first emit
    file_handler = logging.FileHandler(log_file, mode="w", delay=True)
    file_handler.setLevel(level)

    # Create formatter and add it to the handler
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(formatter)

    logger.addHandler(file_handler)

    return logger


@contextmanager
def measure_time():
    start_time = time.perf_counter()
    result = {}
    try:
        yield result
    finally:
        end_time = time.perf_counter()
        result["elapsed_time"] = (end_time - start_time) * 1000


class PydanticDictWrapper(BaseModel):
    """This class creates a pydantic model from a dict object that can be used in the pre_process method of LLMTaskWorker."""

    data: Dict[str, Any]

    def model_dump_json(self, **kwargs):
        return json.dumps(self.data, **kwargs)



# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import sys
import threading
import time
from collections import defaultdict, deque
from queue import Empty, Queue
from threading import Event, Lock
from typing import (
    TYPE_CHECKING,
    DefaultDict,
    Deque,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
)

from .task import Task, TaskWorker
from .web_interface import is_quit_requested, run_web_interface

if TYPE_CHECKING:
    from .graph import Graph


# Type aliases
TaskID = int
TaskName = str
ProvenanceChain = Tuple[Tuple[TaskName, TaskID], ...]


class NotificationTask(Task):
    pass


class Dispatcher:
    def __init__(self, graph: "Graph", web_port=5000):
        self.graph = graph
        self.work_queue = Queue()
        self.provenance: DefaultDict[ProvenanceChain, int] = defaultdict(int)
        self.provenance_trace: Dict[ProvenanceChain, dict] = {}
        self.notifiers: DefaultDict[ProvenanceChain, List[TaskWorker]] = defaultdict(
            list
        )
        self.provenance_lock = Lock()
        self.notifiers_lock = Lock()
        self.stop_event = Event()
        self.active_tasks = 0
        self.task_completion_event = threading.Event()
        self.web_port = web_port
        self.debug_active_tasks: Dict[int, Tuple[TaskWorker, Task]] = {}
        self.completed_tasks: Deque[Tuple[TaskWorker, Task, str]] = deque(
            maxlen=100
        )  # Keep last 100 completed tasks
        self.failed_tasks: Deque[Tuple[TaskWorker, Task]] = deque(
            maxlen=100
        )  # Keep last 100 failed tasks
        self.total_completed_tasks = 0
        self.total_failed_tasks = 0
        self.task_id_counter = 0
        self.task_lock = threading.Lock()

    def _generate_prefixes(self, task: Task) -> Generator[Tuple, None, None]:
        provenance = task._provenance
        for i in range(1, len(provenance) + 1):
            yield tuple(provenance[:i])

    def _add_provenance(self, task: Task):
        for prefix in self._generate_prefixes(task):
            with self.provenance_lock:
                self.provenance[prefix] = self.provenance.get(prefix, 0) + 1
                logging.debug(
                    "+Provenance for %s is now %s", prefix, self.provenance[prefix]
                )
                if prefix in self.provenance_trace:
                    trace_entry = {
                        "worker": task._provenance[-1][0],
                        "action": "adding",
                        "task": task.name,
                        "count": self.provenance[prefix],
                        "status": "",
                    }
                    self.provenance_trace[prefix].append(trace_entry)
                    logging.info(
                        "Tracing: add provenance for %s: %s", prefix, trace_entry
                    )

    def _remove_provenance(self, task: Task):
        to_notify = set()
        for prefix in self._generate_prefixes(task):
            with self.provenance_lock:
                self.provenance[prefix] -= 1
                logging.debug(
                    "-Provenance for %s is now %s", prefix, self.provenance[prefix]
                )

                effective_count = self.provenance[prefix]
                if effective_count == 0:
                    del self.provenance[prefix]
                    to_notify.add(prefix)

                if prefix in self.provenance_trace:
                    # Get the list of notifiers for this prefix
                    with self.notifiers_lock:
                        notifiers = [n.name for n in self.notifiers.get(prefix, [])]

                    status = (
                        "will notify watchers"
                        if effective_count == 0
                        else "still waiting for other tasks"
                    )
                    if effective_count == 0 and notifiers:
                        status += f" (Notifying: {', '.join(notifiers)})"

                    trace_entry = {
                        "worker": task._provenance[-1][0],
                        "action": "removing",
                        "task": task.name,
                        "count": effective_count,
                        "status": status,
                    }
                    self.provenance_trace[prefix].append(trace_entry)
                    logging.info(
                        "Tracing: remove provenance for %s: %s", prefix, trace_entry
                    )

                if effective_count < 0:
                    error_message = f"FATAL ERROR: Provenance count for prefix {prefix} became negative ({effective_count}). This indicates a serious bug in the provenance tracking system."
                    logging.critical(error_message)
                    print(error_message, file=sys.stderr)
                    sys.exit(1)

        for prefix in to_notify:
            self._notify_task_completion(prefix)

    def trace(self, prefix: ProvenanceChain):
        logging.info(f"Starting trace for {prefix}")
        with self.provenance_lock:
            if prefix not in self.provenance_trace:
                self.provenance_trace[prefix] = []

    def watch(
        self, prefix: ProvenanceChain, notifier: TaskWorker, task: Optional[Task] = None
    ) -> bool:
        """
        Watches the given prefix and notifies the specified notifier when the prefix is no longer tracked
        as part of the provenance of all tasks.

        This method sets up a watch on a specific prefix in the provenance chain. When the prefix is
        no longer part of any task's provenance, the provided notifier will be called with the prefix
        as an argument. If the prefix is already not part of any task's provenance, the notifier may
        be called immediately.

        Parameters:
        -----------
        prefix : ProvenanceChain
            The prefix to watch. Must be a tuple representing a part of a task's provenance chain.

        notifier : TaskWorker
            The object to be notified when the watched prefix is no longer in use.
            Its notify method will be called with the watched prefix as an argument.

        task : Task
            The task associated with this watch operation if it was called from consume_work.

        Returns:
        --------
        bool
            True if the notifier was successfully added to the watch list for the given prefix.
            False if the notifier was already in the watch list for this prefix.

        Raises:
        -------
        ValueError
            If the provided prefix is not a tuple.
        """
        if not isinstance(prefix, tuple):
            raise ValueError("Prefix must be a tuple")

        added = False
        with self.notifiers_lock:
            if notifier not in self.notifiers[prefix]:
                self.notifiers[prefix].append(notifier)
                added = True

        if task is not None:
            should_notify = False
            with self.provenance_lock:
                if self.provenance.get(prefix, 0) == 0:
                    should_notify = True

            if should_notify:
                self._notify_task_completion(prefix)

        return added

    def unwatch(self, prefix: ProvenanceChain, notifier: TaskWorker) -> bool:
        if not isinstance(prefix, tuple):
            raise ValueError("Prefix must be a tuple")
        with self.notifiers_lock:
            if prefix in self.notifiers and notifier in self.notifiers[prefix]:
                self.notifiers[prefix].remove(notifier)
                if not self.notifiers[prefix]:
                    del self.notifiers[prefix]
                return True
        return False

    def _notify_task_completion(self, prefix: tuple):
        to_notify = []
        with self.notifiers_lock:
            for notifier in self.notifiers[prefix]:
                to_notify.append((notifier, prefix))

        for notifier, prefix in to_notify:
            logging.info(f"Notifying {notifier.name} that prefix {prefix} is complete")
            with self.task_lock:
                self.active_tasks += 1

            # Use a named function instead of a lambda to avoid closure issues
            def task_completed_callback(future, worker=notifier):
                self._task_completed(worker, None, future)

            future = self.graph._thread_pool.submit(notifier.notify, prefix)
            future.add_done_callback(task_completed_callback)

    def _dispatch_once(self) -> bool:
        try:
            worker, task = self.work_queue.get(timeout=1)
            with self.task_lock:
                self.active_tasks += 1
            future = self.graph._thread_pool.submit(self._execute_task, worker, task)

            # Use a named function instead of a lambda to avoid closure issues
            def task_completed_callback(future, worker=worker, task=task):
                self._task_completed(worker, task, future)

            future.add_done_callback(task_completed_callback)
            return True

        except Empty:
            return False

    def dispatch(self):
        while True:
            # making sure that we can access active_tasks in a thread-safe way
            with self.task_lock:
                if (
                    self.stop_event.is_set()
                    and self.work_queue.empty()
                    and self.active_tasks == 0
                ):
                    break
            self._dispatch_once()

    def _execute_task(self, worker: TaskWorker, task: Task):
        task_id = self._get_next_task_id()
        with self.task_lock:
            self.debug_active_tasks[task_id] = (worker, task)

        try:
            # since we are storing a lot of references to the task, we need to make sure
            # that we are not storing the same task object in multiple places
            worker._pre_consume_work(task.copy())
        except Exception:
            raise  # Re-raise the caught exception
        finally:
            with self.task_lock:
                if task_id in self.debug_active_tasks:
                    del self.debug_active_tasks[task_id]

    def _get_next_task_id(self) -> int:
        with self.task_lock:
            self.task_id_counter += 1
            return self.task_id_counter

    def _task_to_dict(self, worker: TaskWorker, task: Task, error: str = "") -> Dict:
        data = {
            "id": self._get_task_id(task),
            "type": type(task).__name__,
            "worker": worker.name,
            "provenance": [f"{worker}_{id}" for worker, id in task._provenance],
            "input_provenance": [
                {input_task.name: input_task.model_dump()}
                for input_task in task._input_provenance
            ],
        }
        if error:
            data["error"] = error
        return data

    def get_traces(self) -> Dict:
        with self.provenance_lock:
            return self.provenance_trace

    def get_queued_tasks(self) -> List[Dict]:
        return [
            self._task_to_dict(worker, task) for worker, task in self.work_queue.queue
        ]

    def get_active_tasks(self) -> List[Dict]:
        with self.task_lock:
            return [
                self._task_to_dict(worker, task)
                for task_id, (worker, task) in self.debug_active_tasks.items()
            ]

    def get_completed_tasks(self) -> List[Dict]:
        with self.task_lock:
            return [
                self._task_to_dict(worker, task)
                for worker, task in self.completed_tasks
            ]

    def get_failed_tasks(self) -> List[Dict]:
        with self.task_lock:
            return [
                self._task_to_dict(worker, task, error)
                for worker, task, error in self.failed_tasks
            ]

    def _get_task_id(self, task: Task) -> str:
        # Use the last entry in the _provenance list as the task ID
        if task._provenance:
            return f"{task._provenance[-1][0]}_{task._provenance[-1][1]}"
        else:
            # Fallback in case _provenance is empty
            return f"unknown_{id(task)}"

    def _task_completed(self, worker: TaskWorker, task: Optional[Task], future):
        success: bool = False
        error_message: str = ""
        try:
            # This will raise an exception if the task failed
            _ = future.result()

            # Handle successful task completion
            if task:
                logging.info(f"Task {task.name} completed successfully")
            else:
                logging.info(
                    f"Notification for worker {worker.name} completed successfully"
                )
            success = True

        except Exception as e:
            # Handle task failure
            error_message = str(e)
            if task:
                logging.exception(f"Task {task.name} failed with exception: {str(e)}")
            else:
                logging.exception(
                    f"Notification for worker {worker.name} failed with exception: {str(e)}"
                )

            # Anything else that needs to be done when a task fails?

        finally:
            # This code will run whether the task succeeded or failed
            if not success:
                if task is not None:
                    # Determine whether we should retry the task
                    if worker.num_retries > 0:
                        if task.retry_count < worker.num_retries:
                            task.increment_retry_count()
                            with self.task_lock:
                                self.active_tasks -= 1
                                self.work_queue.put((worker, task))
                            logging.info(
                                f"Retrying task {task.name} for the {task.retry_count} time"
                            )
                            return

                with self.task_lock:
                    self.failed_tasks.appendleft(
                        (worker, task if task else NotificationTask(), error_message)
                    )
                    self.total_failed_tasks += 1

                if task:
                    logging.error(
                        f"Task {task.name} failed after {task.retry_count} retries"
                    )
            else:
                with self.task_lock:
                    self.completed_tasks.appendleft(
                        (worker, task if task else NotificationTask())
                    )
                    self.total_completed_tasks += 1

            if task:
                self._remove_provenance(task)

            with self.task_lock:
                self.active_tasks -= 1
                if self.active_tasks == 0 and self.work_queue.empty():
                    self.task_completion_event.set()

    def add_work(self, worker: TaskWorker, task: Task):
        task_copy = task.copy()
        self._add_provenance(task_copy)
        self.work_queue.put((worker, task_copy))

    def add_multiple_work(self, work_items: List[Tuple[TaskWorker, Task]]):
        # the ordering of adding provenance first is important for join tasks to
        # work correctly. Otherwise, caching may lead to fast execution of tasks
        # before all the provenance is added.
        for worker, task in work_items:
            self._add_provenance(task)
        for item in work_items:
            self.work_queue.put(item)

    def stop(self):
        self.stop_event.set()

    def wait_for_completion(self, wait_for_quit=False):
        self.task_completion_event.wait()

        if wait_for_quit:
            while not is_quit_requested():
                # Sleep for a short time to avoid busy waiting
                time.sleep(0.1)

    def start_web_interface(self):
        web_thread = threading.Thread(
            target=run_web_interface, args=(self, self.web_port)
        )
        web_thread.daemon = (
            True  # This ensures the web thread will exit when the main thread exits
        )
        web_thread.start()
